/* gcc -c -O -Wall irq7.c */
/* I took the interupt handling routines from some dos code written 
  by Peter H. Anderson in some WWW page describing the handling of IRQ7. 
  I watched lp.c and plip.c to understand how to write this driver.
*/

/* sending signals to tasks requesting it : i use an linked chain in
which i store all the processes, and so when the signal is sent
i only send to these processes.  */

#define MODULE
#define __KERNEL__ 

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/ioport.h> /* request_region ... */

#include <asm/io.h>        /* outb and inb */
#include <asm/irq.h>       /* enableirq() and disableirq() */
#include <linux/malloc.h>  /* for kalloc() */
#include <signal.h>        /* SIGUSR1 defnied here */

#include "irq7.h"    /* def of ioctl SIGNALS */

#define P_MAJOR 36

struct pidchain 
    {unsigned long tasks;struct pidchain *next;} *linkedchain;

void open_intserv(void);  
void int_processed(void); 
void ip_init(void);
static void irqhandler(int,void *dev_id,struct pt_regs*);

static int irq7_open(struct inode * inode, struct file * file)
{ /*printk("open of irq7 device acknowledged\n");*/
  /* might be better to grab the interrupt here maybe ... dunno */
  /* ret=request_irq(value,&irqhandler,SA_INTERRUPT,"irq7");
     ( ret is int; and value is unsigned int; )*/
 return(0);
}

static int irq7_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
                     unsigned long arg)
{switch (cmd) 
 {case REGISTER_PID: 
  {struct pidchain *which_pid;
   printk("process %ld has registered irq7\n",arg);
   if (linkedchain==NULL) 
   {linkedchain=(struct pidchain*)kmalloc(sizeof(struct pidchain),GFP_KERNEL);
    which_pid=linkedchain;}
   else {which_pid = linkedchain;
    while (which_pid->next !=NULL) which_pid=which_pid->next;
    which_pid->next=(struct pidchain*)kmalloc(sizeof(struct pidchain),GFP_KERNEL); 
    which_pid=which_pid->next;}
   which_pid->next=NULL;
   which_pid->tasks=arg;
  /* uncomment if u wanna see the registered processes in /usr/adm/messages */
  /* which_pid=linkedchain;while(which_pid!=NULL) 
     do {printk("%d",which_pid->tasks); which_pid=which_pid->next;} 
     while(which_pid!=NULL) ; printk("\n");*/
	              }
   break;
   case QUIT_LIST: { struct pidchain *which_pid, *tmp;
       which_pid=linkedchain;
 /* if the leaving process is the 1st one to have registered ... */
       if (which_pid->tasks==arg) {tmp=linkedchain->next;
	  printk("process %ld quit",arg);
          kfree(linkedchain);linkedchain=tmp;return(0);}
 /* ... or if it is in the middle of the chain */
      else do { 
        if (which_pid->next==NULL) return(-1); /*process hadnt registered*/
           else if (which_pid->next->tasks==arg)
              {tmp=which_pid->next->next;kfree(which_pid->next);
	       printk("process %ld quit",arg);
               which_pid->next=tmp;return(0);}
              else which_pid=which_pid->next;
              } while (which_pid!=NULL); 
            printk("Error in linked chain in irq7.c") ;
        }
	break;
	  default : return(-1); /* here is where to add additional commands */
	break;
  }
  return(0);
}

static struct file_operations lp_fops = 
 { NULL, NULL, NULL, NULL, NULL, irq7_ioctl, NULL, irq7_open, NULL,NULL,NULL,};

/* Uses interrupt service routine to note interrupt from printer port.  
   The interrupt is caused by a negative on /ACK input on Printer Port. */
	 
#define DATA 0x0378  
#define STATUS DATA+1  
#define CONTROL DATA+2  

#define TRUE 1  
#define FALSE 0  
	 
/* function called when interrupt is detected */ 
static void irqhandler(int irq, void *dev_id,struct pt_regs * regs)
{struct pidchain *which_pid;
 disable_irq(7);
 int_processed();
 printk("IRQ7 ACTIVATED \n");
 which_pid=linkedchain; 
 if (which_pid!=NULL) /* ie someone has requested the interrupt signal */
 do {kill_proc(which_pid->tasks,SIGUSR1,1);
     which_pid=which_pid->next;} while (which_pid!=NULL);
/* for_each_task(p) if (p->pid==wanted) {send_sig(SIGUSR1,p,1);}*/
/* cannot do that because a process which receives a SIGNAL it is
   not designed to handle will exit                               */
enable_irq(7);
} 

void open_intserv(void)  
/* enables IRQ7 interrupt.  On interrupt (low on /ACK) jumps to intserv.  
   all interrupts disabled during this function; enabled on exit.  */
{  
  int int_mask;  
  int_mask=inb(0x21);           /* 1101 1111 */  
  outb(int_mask & ~0x80,0x21);  /* set bit 7 to zero, leave others alone */  
} 

void int_processed(void)  /* signals 8259 in PC that int has been processed */  
{outb(0x20, 0x20);}  

int init_module(void) 
{int err,irqval;
 linkedchain=NULL;
 if ((err=check_region(DATA,3))<0) 
   {printk("error in check_region()\n");return err;} /* busy */
 request_region(DATA,3,"irq7"); 
 /* the device must have been created previously with 
     mknod /dev/adc c 6 5       (major=6, minor=5) */
 err=register_chrdev(P_MAJOR,"irq7",&lp_fops);
 if (err<0) {printk("unable to get major %d for irq7\n", P_MAJOR);
     release_region(DATA,3);}
     else printk("irq7 module loaded : major=%d\n", P_MAJOR);
 open_intserv(); /* set mask to activate irq7 (interrupt controller) */ 
 outb(inb(CONTROL) | 0x10,CONTROL); /* same thing on parallel port */ 
   /* set bit 4 on control port (irq enable) to logic one */  
  irqval=request_irq(7,irqhandler,SA_INTERRUPT,"irq7",NULL); 
   /* SA_INTERRUPT = shared interrupt ? taken from a post by C. Beauregard 
   (KHG Discussion Page) */
 return 0; 
}

void cleanup_module(void){
unregister_chrdev(P_MAJOR,"irq7");
release_region(DATA,3);printk("irq7 released\n");}

